(* Le code écrit en cours, le code à compléter est 
   à partir de la ligne 60. :*)

(* Création d'un échiquier vide *)
let create_board n =
  Array.init n (fun _ -> Array.make n false)

(* Dans l'échiquier donné, est que la colonne au dessus de (row, col) est libre *)
let rec free_column board row col =
  if row < 0 then true
  else
    not (board.(row).(col)) &&
    free_column board (row-1) col

(* Dans l'échiquier donné, est que la diagonale gauche au dessus
   de (row, col) est libre.  *)

let rec free_diagonal_left board row col =
  if row < 0 || col < 0 then true
  else
    not (board.(row).(col)) &&
    free_diagonal_left board (row-1) (col-1)

(* Dans l'échiquier donné, est que la diagonale droite au dessus
   de (row, col) est libre.  *)

let rec free_diagonal_right board row col n =
  if row < 0 || col >= n then true
  else
    not (board.(row).(col)) &&
    free_diagonal_right board (row-1) (col+1) n

(* Recherche d'une solution avec retour arrière *)
let nqueens f n =
  (* num: nombre de reines posées *)
  let board = create_board n in
  let rec search row =
    if n = row then
      (* solution trouvée *)
      f board
    else begin
      (* solution pas trouvée *)
      for k = 0 to n - 1 do
        if free_column board row k &&
           free_diagonal_left board row k &&
           free_diagonal_right board row k n then
          begin
            board.(row).(k) <- true;
            search (row+1);
            (* ici, le retour arrière, on revient de l'appel
               récursif, on restore l'état précédent et
               on continue la boucle for, i.e. on continue
               de chercher.*)
            board.(row).(k) <- false;
          end
      done
    end
  in
  search 0

(* 1.1 *)
let compatible r1 c1 r2 c2 =
  r1 != r2 (* pas la même ligne *) 
  && c1 != c2 (* pas la même colonne *)
  && abs (r1 - r2) != abs (c1 - c2) (* pas la même diagonale *)

(* 1.2 *)
let rec compatible_list row col other_queens =
  match other_queens with
    [] -> true
  | (ri, ci) :: ll -> compatible row col ri ci && compatible_list row col ll

(* 1.3 *)
let nqueens_list f n =
  let rec search row other_queens =
    if n = row then
      (* solution trouvée *)
      f other_queens
    else begin
      (* solution pas trouvée *)
      for k = 0 to n - 1 do
        if compatible_list row k other_queens then
          search (row+1) ((row, k)::other_queens)
      done
    end
  in
  search 0 []


let time f arg =
  Gc.full_major (); (* on force le GC pour éviter (sans garantie)
                       qu'il s'éxécute pendant notre fonction.*)
  Gc.compact ();
  let t0 = Unix.gettimeofday () in
  let r = f arg in
  let t1 = Unix.gettimeofday () in
  r, (t1 -. t0) *. 1000.

(* 1.4 *)
let main () =
  for i = 0 to 14 do
    let count = ref 0 in
    let _, t = time (nqueens (fun _board -> incr count)) i in
    Printf.printf "Original: %d -> %d (%fms)\n%!" i !count t;
    count := 0; (* remettre la référence à 0 *)
    let _, t = time (nqueens_list (fun _board -> incr count)) i in
    Printf.printf "List: %d -> %d (%fms)\n\n%!" i !count t;
  done

let () = main () (* commenter pour les questions suivantes *)

(* 1.5 La fonction print_list est dans le fichier utils.ml *)

(* 1.6 *)
let test_print_4 () =
  nqueens_list (fun b -> Utils.print_list b 4) 4

let () = test_print_4 () (* commenter pour les questions suivantes *)


(* 2.1 *)
exception Found of (int * int) list

(* On définit une fonction auxiliaire qui rattrape l'exception *)
let nqueens_first n =
  try
    nqueens_list (fun b -> raise (Found b)) n;
    [] (* ici l'exception n'a pas été levée, donc
          pas de solution.*)
  with
    Found b -> b

let search_one_solution () =
  for n = 0 to 100 do (* interrompre le programme avec CTRL-c vers n = 30 *)
    let _, t = time nqueens_first n in
    Printf.printf "N=%d -> %fms\n%!" n t
  done
(*
let () = search_one_solution () (* commenter pour les questions suivantes *)
*)
(* 2.2 *)
let nqueens_random n =
  let rec search row other_queens =
    if n = row then
      (* solution trouvée *)
      Some other_queens
    else begin
      (* solution pas trouvée *)
      let n_pos = ref 0 in
      let pos = ref [] in
      for k = 0 to n - 1 do
        if compatible_list row k other_queens then
          begin
            n_pos := !n_pos + 1;
            pos := (row, k)::!pos;
          end
      done;
      if !n_pos = 0 then None else
        let k = Random.int !n_pos in
        search (row+1) ((List.nth !pos k)::other_queens)
    end
  in
  search 0 []

let search_random_solution n =
  let rec loop tries =
    match nqueens_random n with
      None -> loop (tries+1)
    | Some b -> b, tries
  in
  if n = 2 || n = 3 then [], 0
  else loop 0

(* 2. 3 *)
let test_random () =
  for n = 0 to 100 do
    let (_,tries), t = time search_random_solution n in
    Printf.printf "Random: %d -> %d essais (%fms)\n%!" n tries t;
  done
let () = test_random ()